/**
 * Pyx4me framework
 * Copyright (C) 2006-2008 pyx4j.com.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 * 
 * @author vlads
 * @version $Id: ObexBluetoothClient.java 2796 2008-12-02 23:24:31Z vlads $
 */
package com.pyx4me.maven.obex;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.OutputStream;

import javax.bluetooth.LocalDevice;
import javax.microedition.io.Connector;
import javax.obex.ClientSession;
import javax.obex.HeaderSet;
import javax.obex.Operation;
import javax.obex.ResponseCodes;

import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugin.logging.Log;

public class ObexBluetoothClient {

	private static class ConsoleProgressMonitor {

		long total;

		long startedTime;

		long complete;

		long printedTime;

		static final long STEP_COUNT = 15;

		static final long STEP_INTERVAL = 2 * 1000;

		ConsoleProgressMonitor(long total) {
			this.startedTime = System.currentTimeMillis();
			this.complete = 0;
			this.total = total;
			this.printedTime = 0;
		}

		void transferProgress(int sent) {
			this.complete += sent;
			long now = System.currentTimeMillis();
			if ((printedTime == 0) || ((now - printedTime) > STEP_INTERVAL)) {
				StringBuffer b = new StringBuffer();
				b.append("[INFO] Transferring: ");
				b.append(complete / 1024).append("/").append(total / 1024).append("K");

				b.append(" [");
				for (int i = 0; i < STEP_COUNT; i++) {
					long eta = (i * total) / STEP_COUNT;
					if (eta > complete) {
						b.append(" ");
					} else {
						b.append(".");
					}
				}
				b.append("] ");
				b.append((long) (100 * complete / total)).append("%");

				b.append("\r");
				System.out.print(b.toString());
				printedTime = now;
			}
		}

		void transferComplete(Log log) {
			if (printedTime != 0) {
				long sec = (System.currentTimeMillis() - this.startedTime) / 1000;
				log.info("Transfer " + (total / 1024) + "K completed in " + sec + " sec                     ");
			}
		}
	}

	public static void obexPut(File jar, int connectionRetry, ServerConfig cfg, Log log) throws MojoFailureException {
		AbstractDeviceMojo.configureBlueCove(cfg, log);
		ClientSession clientSession = null;
		FileInputStream is = null;
		try {
			// System.setProperty("bluecove.debug", "true");
			String serverURL;

			if (cfg.url.startsWith("tcpobex://")) {
				serverURL = cfg.url;
			} else if (cfg.url.startsWith("btgoep://")) {
				serverURL = cfg.url;
			} else {
				String url = cfg.url;
				if (url.indexOf(':') == -1) {
					url = ServiceSearchMojo.findOBEX(url, cfg.alias, log);
					if (url == null) {
						throw new MojoFailureException("OBEX Service not found");
					}
				} else if ("bluesoleil".equals(LocalDevice.getProperty("bluecove.stack"))) {
					// in BlueSoleil You need to run
					// DiscoveryAgent.searchServices() with UUID of your service
					// before you can make connection to it using
					// Connector.open("btspp://...");
					int portStart = url.indexOf(':');
					if (portStart != -1) {
						url = ServiceSearchMojo.findOBEX(url.substring(0, portStart), cfg.alias, log);
						if (url == null) {
							throw new MojoFailureException("OBEX Service not found");
						}
					}
				}
				serverURL = "btgoep://" + url;
			}

			log.info(cfg.alias + " Connecting ...");
			int connectionOpenTry = 0;
			while (clientSession == null) {
				try {
					clientSession = (ClientSession) Connector.open(serverURL);
				} catch (IOException e) {
					connectionOpenTry++;
					if (connectionOpenTry > connectionRetry) {
						throw e;
					}
					Thread.sleep(500);
					log.warn(cfg.alias + " connection retry after error " + e.getMessage());
				}
			}
			HeaderSet hsc = clientSession.connect(clientSession.createHeaderSet());
			log.debug("connect responseCode " + hsc.getResponseCode());

			HeaderSet hsp = clientSession.createHeaderSet();
			hsp.setHeader(HeaderSet.NAME, jar.getName());
			hsp.setHeader(HeaderSet.TYPE, "jar");
			long jarLength = jar.length();
			hsp.setHeader(HeaderSet.LENGTH, new Long(jarLength));

			log.info(cfg.alias + " Sending " + jar.getName());
			Operation po = clientSession.put(hsp);

			ConsoleProgressMonitor progress = new ConsoleProgressMonitor(jarLength);
			OutputStream os = po.openOutputStream();

			is = new FileInputStream(jar);
			byte[] buffer = new byte[0x400];
			int i = is.read(buffer);
			while (i != -1) {
				os.write(buffer, 0, i);

				// Show progress
				progress.transferProgress(i);

				i = is.read(buffer);
			}

			os.close();

			log.debug("put responseCode " + po.getResponseCode());

			po.close();

			progress.transferComplete(log);

			HeaderSet hsd = clientSession.disconnect(null);
			log.debug("disconnect responseCode " + hsd.getResponseCode());

			if (hsd.getResponseCode() == ResponseCodes.OBEX_HTTP_OK) {
				log.info(cfg.alias + " Finished successfully");
			}

		} catch (IOException e) {
			log.error("Communication error", e);
			throw new MojoFailureException("Can't upload file");
		} catch (Throwable e) {
			log.error("Error", e);
			throw new MojoFailureException("Can't upload file");
		} finally {
			if (clientSession != null) {
				try {
					clientSession.close();
				} catch (IOException ignore) {
				}
			}
			if (is != null) {
				try {
					is.close();
				} catch (IOException ignore) {
				}
			}
			clientSession = null;
		}
	}
}
